use Getopt::Std;

require conv_check;
require conv_generate;
require conv_parse;

########### Global Definitions:

$debug_msg = 0         ;            # If 1 then print debugging messages
$disable_model_error_checking = 0;  # If 1 then disables error check on the model
$revision = "1.0";                  # Revision number
$input_file = "none";               # filename of the input file
$output_file = "none";              # filename of the output file

########### State information

$defined_SPEX_report = 0;
$defined_SPEX_convergence = 0;
$cur_model_parse_state = "none";
$verify_mode = 0;

$cur_line = 0;
$cur_processor = "none";
$cur_task = "none";
$cur_entry = "none";
$cur_activity = "none";
$cur_number_of_phases = -1;
$op_mode = "none";


########### Variables that contain data retrieved from input file

$model_comments = "";
$model_pragmas = "";
$iteration_limit = 100;
$convergence_limit = 0.00001;
$print_interval = 1;
$underrelaxation_coeff = 0.9;
$solver = "";
$solver_parameters = "";

$SPEX_init_block = "";
$SPEX_report_block = "";
$SPEX_convergence_block = "";

$SPEX_obs_var_list = "";
$SPEX_control_var_list ="";

%processor_list = ();
%task_list1 = ();
%task_list2 = (); 

%entry_list_priority = ();
%entry_list_arriv_rate = ();
%entry_list_service = ();
%entry_list_maxservice = ();
%entry_list_COV = ();
%entry_list_phasetype = ();
%entry_list_thinktime = ();
%entry_list_activities = ();

%call_list_number = ();
%call_list_fanin = ();
%call_list_fanout = ();
%call_list_forwarding = ();

%activity_list_service = ();
%activity_list_COV = ();
%activity_list_phasetype = ();
%activity_list_thinktime = ();
%activity_connections = ();
%activity_calls = ();
%activity_list_entry = ();  # The reverse of entry_list_activities, only present in conv_spread
%task_activity_lookup_list = ();    # Used for error checking only

sub start_program
{
#get the command line options before we do anything else

   &get_command_line_options;
   &open_input_file($input_file);

   @start_tokens = &get_next_line(0);

   if (@start_tokens[0] eq "G")
   {
      $op_mode="LQN";

      if ($debug_msg > 0)
      {
         print "We are in LQN mode\n";
      }

      &parse_initial_information_LQN;
   }
   else
   {
      $op_mode = "SPEX";

      if ($debug_msg > 0)
      {
         print "We are in SPEX mode\n";
      }

      &parse_initial_information_SPEX(join(" ", @start_tokens));
   }

   @start_tokens = &get_next_line(0);
   $num_start_tokens = @start_tokens;

   &check_valid_integer(@start_tokens[1]);

   if (@start_tokens[0] ne "P")
   {
      die "ERROR on line $cur_line: Expecting P <integer> which denotes the start of a Processor block.\n";
   }

   if ($num_start_tokens != 2)
   {
      die "ERROR on line $cur_line, $#start_tokens, @start_tokens: There seems to be some too little or too much information at the start of the processor block.\n";
   }

   &parse_processor_block;

   @start_tokens = &get_next_line(0);
   $num_start_tokens = @start_tokens;

   if (@start_tokens[0] ne "T")
   {
      die "ERROR on line $cur_line: Expecting T <integer> which denotes the start of a Task block.\n";
   }

   if ($num_start_tokens != 2)
   {
      die "ERROR on line $cur_line: There seems to be some too little or too much information at the start of the task block.\n";
   }

   &parse_task_block;

   @start_tokens = &get_next_line(0);
   $num_start_tokens = @start_tokens;

   if (@start_tokens[0] ne "E")
   {
      die "ERROR on line $cur_line: Expecting E <integer> which denotes the start of a Entry block.\n";
   }

   if ($num_start_tokens != 2)
   {
      die "ERROR on line $cur_line: There seems to be some too little or too much information at the start of the entry block.\n";
   }

   &parse_entry_block;

   @start_tokens = &get_next_line(1);
   $num_start_tokens = @start_tokens;

   $activity_block_present = 1;

   while ($activity_block_present == 1)
   {
      if (@start_tokens[0] eq "A")
      {
	 $activity_block_present = 1;

	 if ($num_start_tokens != 2)
	 {
	    die "ERROR on line $cur_line: There seems to be some too little or too much information at the start of the Activity block.\n";
	 }

	 if (!defined $task_list1{@start_tokens[1]})
	 {
	    die "ERROR on line $cur_line: Activities are being bound to task \"@start_tokens[1]\" which has not been defined.\n";
	 }
	 &parse_activity_block(@start_tokens[1]);
         @start_tokens = &get_next_line(1);
         $num_start_tokens = @start_tokens;
      }
      else
      {
	 $activity_block_present = 0;
      }
   }

   if ($op_mode eq "LQN")
   {
      if ($disable_model_error_checking == 0)
      {
         &check_model_for_errors;
      }
      else
      {
         print "Checking model for errors has been DISABLED\n";
      }

      if ($verify_mode == 0)
      {
         &generate_LQN_spreadsheet($output_file);
      }
   }
   else
   {
# Look for the reporting and convergence blocks, both may not be present

      if ($activity_block_present == 1) 
      {
         @start_tokens = &get_next_line(1);
         $num_start_tokens = @start_tokens;
      }

      if ((@start_tokens[0] ne "R") && ($num_start_tokens != 2))
      {
         print "WARNING - There is no SPEX report block defined.\n";
      }
      elsif ((@start_tokens[0] eq "R") && ($num_start_tokens == 2))
      {
         $last_line = get_next_line_until(0, "\\s*\\-1\$", \$SPEX_report_block);
         $last_line =~ s/\-1//;
         $SPEX_report_block = $SPEX_report_block.$last_line;
      }
      else
      {
         die "ERROR at line $cur_line - Expecting to see the start of the SPEX REPORT block.\n";
      }

      if ((@start_tokens[0] eq "C") && ($num_start_tokens == 2))
      {
         die "ERROR - You cannot define a convergence block before the report block.  Both convergence and report blocks must exist if the convergence block is defined\n";
      }

      @start_tokens = &get_next_line(1);
      $num_start_tokens = @start_tokens;
      $defined_SPEX_convergence = 0;

      if ((@start_tokens[0] ne "C") && ($num_start_tokens == 2))
      {
         die "ERROR at line $cur_line: It looks like a SPEX convergence block is trying to be defined incorrectly.\n";
      }
      elsif ((@start_tokens[0] eq "C") && ($num_start_tokens == 2))
      {
         $defined_SPEX_convergence = 1;
         $last_line = get_next_line_until(0, "\\s*\\-1\$", \$SPEX_convergence_block);
         $last_line =~ s/\-1//;
         $SPEX_convergence_block = $SPEX_convergence_block.$last_line;
      }

      if ($disable_model_error_checking == 0)
      {
         &check_model_for_errors;
      }
      else
      {
         print "Checking model for errors has been DISABLED\n";
      }

      if ($verify_mode == 0)
      {
         &generate_SPEX_spreadsheet($output_file);
      }
   }

   &close_input_file;


   if ($verify_mode == 0)
   {
     if ($op_mode eq "LQN")
      { 
         &generate_LQN_spreadsheet($output_file);
      }
      else
      {
         &generate_SPEX_spreadsheet($output_file);
      }
   }
   else
   {
# If we got this far, then there are no errors in the model file...
       print "No errors found in model.\n";
   }
}

#######################################################

sub get_command_line_options
{
  my @temp;
  my @temp2;
  my $temp;

  $opt_d = 0;
  $opt_h = 0;
  $opt_V = 0;
  $opt_v = 0;
  $opt_e = 0;
  
  getopts('Vvdeh');

  if ($opt_d)
  {
     $debug_msg = 1;
  }

  if ($opt_e)
  {
     $disable_model_error_checking = 1;
  }  

  if ($opt_h)
  {
     print "Usage: convspread [-Vvdeh] input_file [output_file]\n\n";
     print "-V           displays revision number of program.\n";
     print "-v           just verify model and do not generate a file.\n";
     print "-d           displays debugging output while parsing the spreadsheet input file.\n";
     print "-e           disables model error checking.\n";
     print "-h           displays this help message.\n";
     print "input_file   ASCII text output from a spreadsheet in the correct format.\n";
     print "output_file  filename of the LQN or SPEX output file.\n\n";
     exit(1);
  }

  if ($opt_V)
  {
     print "SPEX/LQN to Spreadsheet Interface Parser, Version $revision\n";
     print "Copyright 1998, Real-time and Distributed Systems Group\n";
     print "Department of Systems and Computer Engineering,\n";
     print "Carleton University, Ottawa, Canada.\n\n";
     exit(1);
  }

  if ($opt_v)
  {
     $verify_mode = 1;
  }

  $input_file = shift @ARGV;

  if ($input_file eq "")
  {
     die "ERROR - you must specify at least an input filename.  Use -h for more information on usage of this program.\n";
  }

  if ($#ARGV > -1)
  {
     $temp = shift @ARGV;
     @temp2 = split(/\./, $temp);
     $output_file = @temp2[0];
  }
  else
  {
     @temp = split(/\./, $input_file);
     $output_file = @temp[0];
  }

# At this point $#ARGV is going to be -1...
}



###################################################


sub extract_SPEX_variables
{
# $_[0] variable reference to glob of text which contains the SPEX variables
# $_[1] variable reference a blank array which will contain a list of variables extracted from the
#       glob of text.

   my @rough_var_list_temp = ();
   my @rough_var_list = ();
   my $cur_var;

   @rough_var_list_temp = split(/\$/, ${$_[0]});

   my $num_vars = @rough_var_list_temp;

# element #0 is always going to be blank or have data which is not a variable

   @rough_var_list = @rough_var_list_temp[1 .. ($num_vars-1)];
   $z = 0;

   foreach $cur_var (@rough_var_list)
   {
# find the point where the variable will end, and substitute that point with something we know

      $cur_var =~ s/[\n|,| |\[|\]|*|\/|+|=|\-|\;|\(|\)|\{|\}|\^|\%|\:|\.|<|>|\?]/ZZZxxxZZZ/;

# split the variable de-limiting by the special key we insterted, and we have the variable isolated

      my @var_fin = split("ZZZxxxZZZ", $cur_var);

      @{$_[1]}[$z] = @var_fin[0];
      $z++;
   }
}

# We need to return a TRUE (1), otherwise Perl complains.
1;
